/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "minrtsFftUtil.h"

#ifdef HP
#define finite isfinite
#endif
#define  FFTSIZE 4096
#define  ORDER 4
#define  clight 2.99795e8
#define  PI 3.1415926

/**********************************************************
 * Usage:
 *    oirWaveCal inFile outFile
 *
 *    inFiles is an input file containing "long-scan" fringe data
 *    outFile is the name of a FITS window template file
************************************************************** */


int main (int argc, char** argv) {

   int nRegion, iRegion;
   int maxstep;
   int i,ierr;
   long nRows, iRow, oRow;
   int nx, ny, ix, iy;
   long ipix;
   short *inPix;
   int iKey;

   char outFileName[130];
   char *bang = "!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   imaging_data_table_def *idata, *odata;
   imaging_detector_table_def *inDetectorDef;
   imaging_detector_table_def *outDetectorDef;
   imaging_data_row *inData;

   float *dataStore;
   float *opd;
   float *wavePk;
   float *waveAmp;
   float *work1, *work2, *work3;
   int   *valid, nValid;
   float x;
   float maxAmp;
   int nWave, iWave;
   int ic1, ic2;
   
   float gridParm[3];
   int nfft2;
   minrtsFft1d *ff;

   float crpix;
   float polyCoeff[1+ORDER];
/*
      open input file ; get detector stuff and close
*/
   ierr = 0;
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of %s gave error %i \n",argv[1],
         ierr);
       printf("%s \n",strerror(errno));
      return 1;
   }
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_data_table(inFile, &ierr);
   inData         = make_imaging_data_row(idata);
   nRegion        = idata->nregion;
/*
      set interf channels from nRegion
*/
   if (nRegion == 2) {
      ic1 = 0;
      ic2 = 1;
   } else if (nRegion ==4) {
      ic1 = 1;
      ic2 = 2;
   } else {
      printf(" cant deal with nRegion = %i", nRegion);
      return 1;
   }
/*
      need maxstep for output table definition
*/
   iKey = table_find_key(idata->def, "MAXSTEP"); 
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep, 
      &ierr);
/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[2]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n", outFileName, ierr);
      return 1;
      }
/*    
      copy main header from input to output and close input
*/
   i = fits_movabs_hdu(inFile, 1, 0, &ierr);
   i = fits_copy_header(inFile, outFile, &ierr);
   i = fits_close_file(inFile, &ierr);

/*   
    reopen input file
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of %s gave error %i\n", argv[1], 
      ierr);
      return 1;
   }
   inDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata = open_imaging_data_table(inFile, &ierr);
   i = fits_get_num_rows (inFile,  &nRows, &ierr); 
/* suppress autoscaling to avoid overflow */
   for (i=0; i<idata->nregion; i++) 
      fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, &ierr);
/*
         allocate data structures
*/
      nx = inDetectorDef->naxis[0][0];
      ny = inDetectorDef->naxis[0][1];
      dataStore = (float *)malloc(nRows * nx * sizeof(float));
      opd       = (float *)malloc(nRows * sizeof(float));
      wavePk    = (float *)malloc(nx * sizeof(float));
      waveAmp   = (float *)malloc(nx * sizeof(float));
      work1     = (float *)malloc(nRows * sizeof(float));
      work2     = (float *)malloc(nRows * sizeof(float));
      work3     = (float *)malloc(FFTSIZE * 2 * sizeof(float));
      valid     = (int *)malloc(nRows * sizeof(int));
      nWave  = nx;
      nValid = 0;
      nfft2  = FFTSIZE/2;

/*
         loop over input rows
*/
   for (iRow=1; iRow<=nRows; iRow++) {
      i = read_imaging_data_table(inFile, idata, inData, iRow, &ierr);

/*
            check for invalid time; this indicates EXPREP failure and
            the OPD values are also invalid.  
*/
      if (0 == finite(inData->time)) {
         valid[iRow-1] = 0;
      } else {
         valid[iRow-1] = 1;
         opd[iRow-1] = inData->local_opd[0] - inData->local_opd[1];
/*
            compress data
*/
         for (ix=0; ix<nx; ix++) {
            x = 0;
            inPix = (short *)inData->data[ic1]+ix;
            for (ipix=0;ipix<ny;ipix++) {
               x = x + *(inPix);
               inPix = inPix + nx;
            }
            inPix = (short *)inData->data[ic2]+ix;
            for (ipix=0;ipix<ny;ipix++) {
               x = x - *(inPix);
               inPix = inPix + nx;
            }
            dataStore[ix + nx *(iRow-1)] = x;
         } /* loop over ix */
      }  /* inputdata is valid for this row */
   } /* loop over rows */
/*
        close input data
*/
/*
        remove average from each column of table
*/
   for (ix=0; ix<nx; ix++) {
      nValid = 0;
      x = 0;
      for (iy=0; iy<nRows; iy++) {
         if (valid[iy] ==1) {
            ++nValid;
            x = x + dataStore[ix + nx*iy];
         }
      }
      x = x/nValid;
      for (iy=0; iy<nRows; iy++) {
         if (valid[iy] ==1) {
            dataStore[ix + nx*iy] = dataStore[ix + nx*iy] - x;
         }
      }

   }
/*
      create FFT object
*/
   gridParm[0] = -1.e-6*FFTSIZE/2.;
   gridParm[1] = 1.e-6;
   gridParm[2] = FFTSIZE;

   ff = minrtsMakeFft1d(gridParm, 0);

/*
      loop over observed channels
*/

   for (iWave=0; iWave<nWave; iWave++) {

/*
      copy valid data to work arrays
*/
      iy = -1;
      nValid = 0;
      for (iRow=0; iRow<nRows; iRow++) {
         if (valid[iRow] ==1) {
            ++nValid;
            work1[++iy] = opd[iRow];
            work2[iy] = dataStore[iWave + nx*iRow];
         }
      }
/*
      perform FFT; take absolute values
*/

     minrtsFftGridFft(ff, nValid, work1, work2, 0, work3); 
     for (ix=0;ix<nfft2; ix++) work3[ix] = 
        sqrt(work3[2*ix]*work3[2*ix] + work3[2*ix+1]*work3[2*ix+1]);
/*
      find peak and interpolate
*/
     wavePk[iWave] = -minrtsQPk(nfft2, ff->kGrid, work3, waveAmp+iWave,0);

/*
       convert from wavenumber to frequency
*/
      wavePk[iWave] = wavePk[iWave] * clight/(2.*PI);
   } /* loop over wavelengths */

/*
      do polynomial fit to frequency as function of pixels
      First, only choose points whose amp is 5% of max
*/
   maxAmp = 0;
   for (iWave=0;iWave<nWave;iWave++) 
      if (waveAmp[iWave] > maxAmp) maxAmp=waveAmp[iWave];
   for (iWave=0;iWave<nWave;iWave++) 
      if (waveAmp[iWave] >= 0.05*maxAmp) waveAmp[iWave] = 1.;
         else waveAmp[iWave] = 0.;

/*
      generate pixel positions as float array relative to ref. pixel
      and put in work1
*/
      crpix = inDetectorDef->crpix[0][0];
      for (iWave=0;iWave<nWave;iWave++) work1[iWave] = 1+iWave-crpix;


/*
      do weighted polynomial fit
*/
      minrtsPolyFit(ORDER, nWave, work1, wavePk, waveAmp, polyCoeff);
/*
      insert values into output detector table
         crval = constant value
         cd    = linear value
         dmp/dmc  higher order values normalized to linear value
*/
      for (iRegion=0; iRegion<nRegion; iRegion++) {
         outDetectorDef->crval[iRegion][0] = polyCoeff[0];
         outDetectorDef->cd[iRegion][0] = polyCoeff[1];
         outDetectorDef->cd[iRegion][1] = 0.;
         outDetectorDef->cd[iRegion][2] = 0.;
         outDetectorDef->cd[iRegion][3] = 1.;
/*
              initialize dmp,dmc to zero
*/
         for (i=0;i<outDetectorDef->max_coef;i++) {
            outDetectorDef->dmp[iRegion][i][0][0] = 0;
            outDetectorDef->dmp[iRegion][i][1][0] = 0;
            outDetectorDef->dmp[iRegion][i][0][1] = 0;
            outDetectorDef->dmp[iRegion][i][1][1] = 0;
            outDetectorDef->dmc[iRegion][i][0]   = 0.; 
            outDetectorDef->dmc[iRegion][i][1]   = 0.; 
         }
/*            set x-exponents to 1,2,3...; */

         for (i=0; i<ORDER; i++) outDetectorDef->dmp[iRegion][i][0][0] = 1+i;
         outDetectorDef->dmp[iRegion][0][1][1] = 1;

/*            write coefficients into correct place */

         for (i=0; i<ORDER; i++) outDetectorDef->dmc[iRegion][i][0] = 
            polyCoeff[1+i]/polyCoeff[1];
         outDetectorDef->dmc[iRegion][0][1] = 1;
      } /* iRegion */
/*
      write output detector table
      create output data table (but no data rows)
*/
   odata = make_imaging_data_table_def (outFile, 
      outDetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
      }

/*
      close output stuff
*/
   i = close_imaging_data_table(outFile, odata, &ierr);

   i = fits_close_file(inFile, &ierr);
   kill_imaging_data_row(inData);
   kill_imaging_data_table_def (idata,1);
   minrtsKillFft1d(ff);
   free(dataStore);
   free(opd);
   free(wavePk);
   free(waveAmp);
   free(valid);
   free(work1);
   free(work2);
   free(work3);
   return 0;
}
